(* explicit module definition???????????? *)
(* TODO: check all splitting functions for rounding etc.
  
   REMARK: all intervals passed to this module, are deeply copied

 *)

open Interval

type box = { mutable bounds:interval array ; mutable empty: bool} 

(* constructors *)
val empty_box: unit -> box
val zero_dim: unit -> box  (* the only non-empty 0-dimensional box *)
val full_box: int -> box    (* box of R^{#1} *)
val make: interval array -> box  (* box with bounds #1 *)
val init: int -> (int -> interval) -> box (* box of dimension #1 and k-th element #2(k) *) 
val copy: box -> box  
val equal: box -> box -> bool
val subset: box -> box -> bool   (* non-strict *)
val compare_box :box->box->float->bool (* boolean value #4 is obtained by checking the equality between #1 and #2 depending on the precision value #3*)          
val diff_width_rel: box->box->float->bool (* is the ratio of the width of a component of #2 rel. to #1 < (1 - #3(eps))? *)
val diff_width_rel_vars: box->box->float->int list (* list of variables, for which the ratio of the width of each component of #2 rel. to #1 < (1 - #3(eps)) *)
val diff_vars: box->box->int list (* the list of components for which #1 is not equal #2 *)
val copy_bounds : box -> box ->unit (* copy the bounds of #2 to #1 *)
val copy_box_bounds : box -> box -> int list -> unit

(* access functions *)
val get_bound: box -> int -> interval   (* the #2-th dimension of the box *)
val set_bound: box -> int -> interval -> unit  (* set #2-th dimension of the box #3 *)
val is_empty: box -> bool
val dim: box -> int                         (* dimension of the box, -1 if empty *) 
val width: box-> float
val diagonal: box-> float (* length of diagonal *)
val volume: box -> float
val contains: box -> float array -> bool
val interior: box -> float array -> bool (* is in interior of box *)
val widest_var: box -> int -> int (* widest var from the first #2 coordinates *)
val midpoint_box: box -> box (* midpoint of #1 *) 
val midpoint: box -> float array (* midpoint of #1 *) 
val is_consistent: box -> bool
val intersecting: box -> box -> bool 
val overlapping: box -> box -> bool 

(* destructive functions *)
val set_empty: box -> unit 
val set_full: box -> unit 
val union_d: box -> box -> unit (* #1:= #1 union #2 *)
val cart_prod2_shared: box -> box -> box (* cartesian product, but shares intervals from input boxes *)

(* printing *)
val print: box -> unit
val print_debug: box -> unit (* print in exact represention of boundaries *)
val print_boxlist: box list ->string -> unit (* prints the list of boxes #1 and each box is followed by a descriptive charater #2 (stating true|false|unknown boxes) in the binary format on the console *)
(* val print_var_assoc_ivl: box ->string list -> unit *)


(* box arithmetic *)
val cart_prod2: box -> box -> box 
val cart_prod: box list -> box 
val union2: box -> box -> box
val intersection2: box -> box -> box
val union: box list -> box
val bisect: box -> int -> box * box           (* bisect #1 at variable #2 *)
val bisection_face: box -> int -> box         (* the intersection of the two elements of (bisect b k) *)
val diff: box (* B *) -> box (* D *) -> box list 
  (* a set of boxes that
     * are non-overlapping, and
     * whose union is equal to the closure of the set difference B-D
   *)
val proj: box -> int -> int -> box (* project box #1 to #3 variables starting from variable #2 *)
val proj1: box -> box (* project box for one variable *)
val blowup: box -> int -> box list (* split every side of #1 into #2 pieces; creates #2^{dim(#1)} pieces *)

(* I/O *)
val vrml: box list -> int -> int -> int -> unit
(*val var_assoc_ivl : box -> string list -> (string * interval) list *) (*list in which each bounds of the box #1 is associated with each element of the list of variables used*)
(*val pow:int->int->int *)
val box_precision:box->float->unit (*  bounds of #1 are truncated  to #1*#2 *)   
