(* interval evaluation; current version does not support decomposed constraints! *)

open Interval
open Box
open Constr

let ieval_pred (p: predicateUnary) ({ Interval.lo= lo; Interval.hi= hi }: interval) = 
  match p with
    Equal -> { Bool_intv.lo= (lo>=0.0 || hi<=0.0); Bool_intv.hi= (lo<= 0.0 && hi>=0.0)} | 
    LessEqual -> { Bool_intv.lo= (lo <= 0.); Bool_intv.hi= (hi <= 0.) } | 
    Less -> { Bool_intv.lo= (lo < 0.); Bool_intv.hi= (hi < 0.) } |  
    NotEqual -> { Bool_intv.lo= (lo>=0. && hi<=0.); Bool_intv.hi= (lo<=0.0 || hi>=0.0)} | 
    Greater -> { Bool_intv.lo= (lo >= 0.); Bool_intv.hi= (hi >= 0.) } | 
    GreaterEqual -> { Bool_intv.lo= (lo > 0.); Bool_intv.hi= (hi > 0.) } 

let rec ieval_term (c: constr list) (b: box) (v: variable): interval =
  let rec primDefVar (c: constr list) (v: variable): constr = 
    (* primitive constraint in 'c' defining 'v' *)
    match c with 
      [] -> print_string "Variable not found "; print_int v; assert(false) 
    | first::rest ->
	match first with 
	  Ternary { var33= var3 } -> 
	    if var3=v then first
	    else primDefVar rest v |
	  Binary { var22= var2 } -> 
	    if var2=v then first
	    else primDefVar rest v |
	  Power { var02= varP } ->
	    if varP=v then first
	    else primDefVar rest v | 
	  IsElement { var= var } ->
	    if var=v then first 
	    else primDefVar rest v |
	  _ -> primDefVar rest v in
  if v<(dim b) then get_bound b v 
  else begin
    match primDefVar c v with
      Ternary { predicate3= p; var31= v1; var32= v2; var33= v }-> begin
	let i1= ieval_term c b v1 
	and i2= ieval_term c b v2 in
	match p with 
	  Addition -> addIII i1 i2 |
	  Subtraction -> subIII i1 i2 |
	  Multiplication -> mulIII i1 i2 
      end |
      Binary { predicate2= p; var21= v1; var22= v } -> begin
	let iv= ieval_term c b v1 in
	match p with 
	      UMinus -> { lo= -. iv.hi; hi= -. iv.lo } |
	      Sin -> sinII iv | 
	      Cos -> cosII iv | 
	      Exp -> expII iv | 
	      Tan -> tanII iv |
	      Log -> logII iv |
	      ArcSin -> asinII iv |
	      ArcCos -> acosII iv |
	      ArcTan -> atanII iv |
	      Square -> squareII iv
      end |
      Power { var01= var1; n= power; var02= var2 } -> begin
	let i= ieval_term c b var1 in
	    powII i power 
      end |
      IsElement { var= v; interval= interval } ->
	interval |
	_ -> assert(false) 
  end
  
let rec ieval (c: constr) (b: box): (constr * Bool_intv.bintv) =
  let ieval_atomic (c: constr list) (b: box): Bool_intv.bintv =
    let rec root (c: constr list) =
      match c with
	[] -> assert(false) |
	first::rest -> 
	  match first with
	    Unary { predicate1= p; var1= v } -> (p, v) |
	    _ -> root rest 
    in
  begin
    Prune.deduce_atomic_ctr:= !Prune.deduce_atomic_ctr+1;
    let (root_pred, root_var) = root c in
    let res_term= ieval_term c b root_var in 
    let r= 
      match root_pred with
	Equal -> { Bool_intv.lo= res_term.lo>=0.0 && res_term.hi<=0.0; Bool_intv.hi= res_term.lo<=0.0 && res_term.hi>=0.0 } |
	NotEqual -> { Bool_intv.lo= res_term.lo>0.0 || res_term.hi<0.0; Bool_intv.hi= res_term.lo<0.0 || res_term.hi>0.0 } |
	LessEqual -> { Bool_intv.lo= res_term.hi<=0.0; Bool_intv.hi= res_term.lo<=0.0 } |
	Less -> { Bool_intv.lo= res_term.hi<0.0; Bool_intv.hi= res_term.lo<0.0 } |
	GreaterEqual -> { Bool_intv.lo= res_term.lo>=0.0; Bool_intv.hi= res_term.hi>=0.0 } |
	Greater -> { Bool_intv.lo= res_term.lo>0.0; Bool_intv.hi= res_term.hi>0.0 }
    in begin
      assert(Bool_intv.is_consistent r);
      r
    end
  end in
  match c with
    ExistQuant { qbound= qb; subconstr= sc; qnext_branch= -1 } -> begin
      match sc with
	Conjunction cc -> (c, ieval_atomic cc.branches b) |
	_ -> (c, ieval_atomic [ c ] b)
    end |
    ExistQuant { qbound= qb; subconstr= sc; outer_changed= ochanged; inner_changed= ichanged; qnext_branch= nb } -> 
      let (new_sc, r) = ieval sc (cart_prod2 b (if !Prune.flag_sample then midpoint_box qb else qb)) in
      (ExistQuant { qbound= qb; subconstr= new_sc; outer_changed= ochanged; inner_changed= ichanged; qnext_branch= nb }, r) | 
    UnivQuant { qbound= qb; subconstr= sc; outer_changed= ochanged; inner_changed= ichanged; qnext_branch= nb } -> 
      let (new_sc, r) = ieval sc (cart_prod2 b qb) in
      (UnivQuant { qbound= qb; subconstr= new_sc; outer_changed= ochanged; inner_changed= ichanged; qnext_branch= nb }, r) | 
    Disjunction scl -> 
      let rec eval_disj l = 
	match l with 
	  [] -> ([], Bool_intv.false_intv) |
	  first::rest -> 
	    let (c_first, val_first)= (ieval first b) in 
	      if !Prune.flag_shortcut && Bool_intv.is_true val_first then 
(*	      ((if !flag_remove then l else c_first::l), Bool_intv.true_intv) *)
		(c_first::l, Bool_intv.true_intv) 
	      else 
		let (c_rest, val_rest)= (eval_disj rest) in
		((if !Prune.flag_remove && (Bool_intv.is_false val_first) then c_rest else c_first::c_rest), 
		 Bool_intv.disj val_first val_rest) 
	    in
      let (c, ival)= eval_disj scl.branches
      in (Disjunction { scl with  branches= c }, ival) |
    Conjunction scl -> 
      let rec eval_conj l = 
	match l with 
	  [] -> ([], Bool_intv.true_intv) |
	  first::rest -> 
	    let (c_first, val_first)= (ieval first b) in
	    if !Prune.flag_shortcut && Bool_intv.is_false val_first then 
(*	      ((if !flag_remove then l else c_first::l), Bool_intv.true_intv) *)
	      (c_first::l, Bool_intv.true_intv) 
	    else let (c_rest, val_rest)= (eval_conj rest) in
	    ((if !Prune.flag_remove && (Bool_intv.is_true val_first) then c_rest else c_first::c_rest), 
             Bool_intv.conj val_first val_rest) in
      let (c, ival)= eval_conj scl.branches
      in (Conjunction { scl with branches= c }, ival) |
    _ -> assert(false)


