(* Currently we assume that smathlib has been compiled with 
     #define ROUNDING_METHOD FPSET
   in the file sound_ops.c

   We could gain additional speed (5% as of 2012) by using method FAST_FPSET instead.
   But this would entail that the rounding mode has to be set to DOWN before all calls to 
   smathlib, and to NEAREST before all UNIX system calls (LP solvers etc.). 
   This seems to be too risky for gaining 5% 
*) 

type interval={mutable lo:float;mutable hi:float};; 

external narrow_add: interval -> interval -> interval -> bool = "narrow_add_" ;;
external narrow_mul: interval -> interval -> interval -> bool = "narrow_mul_" ;; 
external narrow_square: interval -> interval -> bool = "narrow_square_" ;;
external narrow_u_minus: interval -> interval  -> bool = "narrow_u_minus_" ;;
(* external narrow_and: interval -> interval -> interval -> bool = "narrow_and" ;;
external narrow_or: interval -> interval -> interval -> bool = "narrow_or_" ;;
external narrow_imp: interval -> interval -> interval -> bool = "narrow_imp_" ;;
external narrow_if: interval  -> interval -> interval -> bool = "narrow_if_" ;;
external narrow_xor: interval -> interval -> interval -> bool = "narrow_xor_" ;;
external narrow_not: interval  -> interval -> bool = "narrow_not_" ;;*)
external narrow_pow_odd: interval -> interval -> interval -> bool = "narrow_pow_odd_" ;;
external narrow_pow_even: interval -> interval -> interval -> bool = "narrow_pow_even_" ;;
external narrow_le: interval  -> bool = "narrow_le_" ;;
external narrow_ne: interval -> bool = "narrow_ne_" ;;
external narrow_lt: interval  -> bool = "narrow_lt_" ;;
external narrow_eq: interval -> bool = "narrow_eq_" ;;
external narrow_gt: interval -> bool = "narrow_gt_";;
external narrow_ge: interval -> bool = "narrow_ge_";;
(*external narrow_bool: interval -> interval -> bool = "narrow_bool_" ;;
external narrow_integer: interval -> interval -> bool = "narrow_integer_" ;;
external narrow_abs: interval -> interval -> bool = "narrow_abs_" ;;
external narrow_ceil: interval -> interval -> bool = "narrow_ceil_" ;;
external narrow_flr: interval -> interval -> bool = "narrow_flr_" ;;
external narrow_sgn: interval -> interval -> bool = "narrow_sgn_" ;;
external narrow_subset: interval -> interval -> bool = "narrow_subset_" ;;*)
external narrow_exp: interval -> interval  -> bool = "narrow_exp_" ;;
external narrow_log: interval -> interval  -> bool = "narrow_log_" ;;
external narrow_sin: interval -> interval -> bool = "narrow_sin_" ;;
external narrow_cos: interval -> interval -> bool = "narrow_cos_" ;;
external narrow_tan: interval -> interval -> bool = "narrow_tan_" ;;
external narrow_asin: interval -> interval -> bool = "narrow_asin_" ;;
external narrow_acos: interval -> interval -> bool = "narrow_acos_" ;;
external narrow_atan: interval -> interval -> bool = "narrow_atan_" ;;
(*external narrow_max: interval -> interval -> interval -> bool = "narrow_max_" ;;
external narrow_min: interval -> interval -> interval -> bool = "narrow_min_" ;;*)
external addIII: interval -> interval ->interval = "addIII_";;
external subIII : interval ->interval ->interval = "subIII_";;
external mulIII : interval ->interval ->interval = "mulIII_";;
external divIII : interval ->interval ->interval = "divIII_";;
external intersectIII : interval ->interval ->interval = "intersectIII_";;
external unionIII : interval ->interval ->interval = "unionIII_";;
external maxIII : interval ->interval ->interval = "maxIII_";;
external minIII : interval ->interval ->interval = "minIII_";;
external negII : interval ->interval = "negII_";;
external sqrtII : interval ->interval = "sqrtII_";;
external squareII : interval ->interval = "squareII_";;
external absII : interval ->interval = "absII_";;
external sgnII : interval ->interval = "sgnII_";;
external floorII : interval ->interval = "floorII_";;
external ceilII : interval ->interval = "ceilII_";;
external two_to_nDI : int ->interval = "two_to_nDI_";;
external expII : interval ->interval = "expII_";;
external logII : interval ->interval = "logII_";;
external sinII : interval ->interval = "sinII_";;
external cosII : interval ->interval = "cosII_";;
external tanII : interval ->interval = "tanII_";;
external asinII : interval ->interval = "asinII_";;
external acosII : interval ->interval = "acosII_";;
external atanII : interval ->interval = "atanII_";;

external next_fp : float -> float = "next_fp_";;
external prev_fp : float -> float = "prev_fp_";;

external get_rounding: unit -> int = "get_rounding_";; 
external set_rounding: int -> unit = "set_rounding_";; 
external set_rounding_downward: unit -> unit = "set_rounding_downward_";; 
external set_rounding_upward: unit -> unit = "set_rounding_upward_";; 
external set_rounding_tonearest: unit -> unit = "set_rounding_tonearest_";; 
external set_rounding_towardzero: unit -> unit = "set_rounding_towardzero_";; 
external print_float_exact_extern: float -> unit = "print_float_exact_";;

let print_float_exact (f: float): unit =
  flush stdout;
  print_float_exact_extern f


let float_to_intv (a: float): interval = { lo= a; hi= a }

let int_to_intv (n: int): interval = float_to_intv (float_of_int n)

let rec powII (iv: interval) (p: int): interval =
  match p with 
    0 -> { lo= 1.0; hi= 1.0 } |
    1 -> iv | 
    2 -> squareII iv |
    n -> if (n mod 2 = 0) then powII (squareII iv) (p / 2) 
         else mulIII (sgnII iv) (expII (mulIII (int_to_intv p) (logII (absII iv))))         (* sgn(x)*exp(y*log(|x|)) *)

let narrow_sub (iv1: interval) (iv2: interval) (iv3: interval) =
  narrow_add iv2 iv3 iv1

(* external narrow-function is doing the job I think (Sven) 
let narrow_atan (iv1: interval) (iv2: interval): bool =
  if iv2.hi>Constants.halfpi_hi then iv2.hi<- Constants.halfpi_hi;
  if iv2.lo< -. Constants.halfpi_hi then iv2.lo<- -. Constants.halfpi_hi;
  narrow_tan iv2 iv1
*)

let copy (iv: interval) =
  { lo= iv.lo; hi= iv.hi }

let empty (): interval = { lo= infinity; hi= neg_infinity }

let everything (): interval = { lo= neg_infinity; hi= infinity }

let is_empty (a: interval): bool = a.hi<a.lo 

let width (a: interval) =  a.hi -. a.lo

let union (iv1: interval) (iv2: interval) = { lo= min iv1.lo iv2.lo; hi= max iv1.hi iv2.hi }

let midpoint (iv: interval) =
 (iv.hi +. iv.lo) /. 2.0

let bisect (iv: interval) =
  ( { lo= iv.lo; hi= midpoint iv},
    { lo= midpoint iv; hi= iv.hi })

let diff (iv1: interval) (iv2: interval): interval list= 
  (List.concat [
    if iv1.lo<iv2.lo then [ { lo= iv1.lo; hi= iv2.lo } ] else [] ;
    if iv2.hi<iv1.hi then [ { lo= iv2.hi; hi= iv1.hi } ] else [] ]: interval list) 

let split (iv: interval) (k: int) =
  let dist= (iv.hi -. iv.lo)/. float_of_int(k) in
  let f (i:int) = { lo= iv.lo +. dist *. (float_of_int i); hi= iv.lo +. dist *. (float_of_int (i+1)) } in
  Array.init k f

let print (iv: interval) = 
  print_string "[ ";
  print_float iv.lo;
  print_string ", ";
  print_float iv.hi;
  print_string "]"

let print_exact (iv: interval) = 
  print_string "[ ";
  print_float_exact iv.lo;
  print_string ", ";
  print_float_exact iv.hi;
  print_string "]"

let print_ivl_bin (iv :interval) =
  let print_int64 (y :int64) =
    for i = 7 downto 0 do
      (output_byte stdout (Int64.to_int (Int64.logand (Int64.shift_right_logical y (8*i)) (Int64.of_int 255)))); 
    done in
  print_int64 (Int64.bits_of_float iv.lo);
  print_int64 (Int64.bits_of_float iv.hi)

let print_debug (iv: interval)= 
  let print_float (f: float) =
    let f= Int64.bits_of_float f in
    let mantissa= Int64.logand f (Int64.of_string "0xfffffffffffff")
    and exponent= Int64.logand (Int64.shift_right_logical f 52) (Int64.of_string "0x7ff") 
    and sign= (Int64.to_int (Int64.shift_right_logical f 63)) in 
    if sign>0 then print_string "-";
    print_string (Int64.to_string (Int64.add (Int64.of_string "0x10000000000000") mantissa)); print_string "*";
    print_string ("2^" ^ (Int64.to_string (Int64.sub (Int64.sub exponent (Int64.of_int 1023)) (Int64.of_int 52)))); print_string " " in
  print_float iv.lo; print_string ", ";
  print_float iv.hi

let equal (iv1 :interval) (iv2 :interval)=
  (iv1.lo = iv2.lo) && (iv1.hi = iv2.hi ) 

let intersecting (iv1: interval) (iv2: interval)=
  max iv1.lo iv2.lo<=min iv1.hi iv2.hi

let overlapping (iv1: interval) (iv2: interval)=
  max iv1.lo iv2.lo<min iv1.hi iv2.hi

let subset (iv1 :interval) (iv2 :interval)=
  (iv1.lo >= iv2.lo) && (iv1.hi <= iv2.hi ) 

let intersect (iv1: interval) (iv2: interval) =
  { lo= max iv1.lo iv2.lo; hi= min iv1.hi iv2.hi }

let precision_adjust (iv :interval) (d:float)=
    {lo=(float_of_int (int_of_float (iv.lo *. d))) /. d;hi = (float_of_int (int_of_float (iv.hi *. d))) /. d }

let precision (iv :interval) (d:float)= 
     {lo=float_of_int (int_of_float (iv.lo *. d)); hi = float_of_int (int_of_float (iv.hi *. d))}

let diff_width_rel (iv1: interval) (iv2: interval) (eps: float)=
  (Util.is_a_number (width iv2)) && (width iv2)<= (1.0 -. eps) *. (width iv1) 

let compare_interval (iv1 :interval) (iv2 :interval) (d:float)=
  let iv11 = copy iv1 in 
  let iv22 = copy iv2 in
  let iv11 = {lo=float_of_int (int_of_float (iv11.lo *. d));hi = float_of_int (int_of_float (iv11.hi *. d))} in 
  let iv22={lo = float_of_int (int_of_float (iv22.lo *. d));hi = float_of_int (int_of_float (iv22.hi *. d))} in 
  (iv11.lo = iv22.lo) && (iv11.hi = iv22.hi ) 

let finite_width (iv:interval) = 
  not ((classify_float iv.lo)=FP_infinite) && 
  not ((classify_float iv.hi)=FP_infinite) &&
  not ((classify_float iv.lo)=FP_nan) && 
  not ((classify_float iv.hi)=FP_nan) 

let is_thin (iv: interval) =
  iv.lo=iv.hi

let contains (iv: interval) (f: float): bool=
  iv.lo <= f && f<=iv.hi
  
