%{ 
    (* variable list is passed through Parser_intern.variables; throughout parsing it contains the list of free variables of the current expression *)

      open Constr
      open Interval
      open Parser_intern

      let original_constraint= true (* create the original constraints *)
      let mean_value_recursive= false (* use the recursive version of the mean value constraints (instead of top-level only) *)
      let mean_value_subset= false (* create the subset versions of the mean value constraint; needed for correctness of recursive version *)

      let trace= false

	  (* number of copies of the free variable block *)
      let free_var_blocks= (if !mean_value then 1 else 0)+(if original_constraint then 1 else 0)
 
      let cur_int_var = ref 1000000000

      type term_info = { 
	  constraints: constr list; 
	  term_var: variable;        (* -1 if not used *)
	  sample_eval_var: variable; (* -1 if not used *)
	  deriv_eval_var: variable; (* -1 if not used *)
	  deriv_vars: variable array (* empty array if not used *)
	}
 
      (* variable numbers used for mean-value constraint *)
      let var_deriv = ref (Array.make 0 (Array.make 0 0)) (* the variable number of each dx/dy where x,y in !variables *)
      let var_x_minus_a= ref (Array.make 0 0) (* x-a *)
      let var_deriv_eval = ref (Array.make 0 0) (* y *)

      (* the mean value constraints itself: f(x)= f(a)+(D f)(y)*(x-a), or its subset version *)
      let add_mean_value (constraints: constr list ref) (term_var: variable) (sample_eval_var: variable) (deriv_vars: variable array): unit = 
	let new_constraints= ref [] in
	for i= 0 to (List.length !variables)-1 do (* f'(x-a) *)
	  cur_int_var:= !cur_int_var+1;
	  new_constraints:= 
            Ternary { predicate3= Multiplication; var31= deriv_vars.(i); var32= !var_x_minus_a.(i); var33= !cur_int_var }::
	    !new_constraints;
	  if i>0 then begin (* add up elements for vector product *)
	    cur_int_var:= !cur_int_var+1;
	    new_constraints:= 
	      Ternary { predicate3= Addition; var31= !cur_int_var-2; var32= !cur_int_var-1; var33= !cur_int_var }::!new_constraints;
	  end
	done;
	if mean_value_subset then begin (* f(x) subset f(a)+f'(y)(x-a) *)
	  cur_int_var:= !cur_int_var+1;
	  new_constraints:= 
	    Ternary { predicate3= Addition; var31= sample_eval_var; var32= !cur_int_var-1; var33= !cur_int_var }::
	    SubSet { subset_var1= term_var; subset_var2= !cur_int_var }::
	    !new_constraints;
	end else
	  new_constraints:= (* f(a)+f'(y)(x-a)=f(x) *)
	    Ternary { predicate3= Addition; var31= sample_eval_var; var32= !cur_int_var; var33= term_var; }::
	    !new_constraints;
	if trace then begin
	  print_endline "(add_mean_value";
	  List.iter print_constr !new_constraints;
	  print_endline "add_mean_value)";
	end;
	constraints:= !new_constraints @ !constraints
	
      (* the part of the mean value constraint that is shared between all levels of term tree: y subseteq a+t(x-a) *)
      let mean_value_shared () =
	let constraints= ref [] in

	cur_int_var:= !cur_int_var+1;
	let var_t= !cur_int_var in
	constraints:=
	  IsElement { var= var_t; interval= { lo= 0.0; hi= 1.0 } }::!constraints;

	for k= 0 to (List.length !variables)-1 do 
	  cur_int_var:= !cur_int_var+1;
	  constraints:= (* t(x-a) *)
            Ternary { predicate3= Addition; var31= !var_x_minus_a.(k); var32= k+(List.length !variables); var33= k}::
	    Ternary { predicate3= Multiplication; var31= var_t; var32= !var_x_minus_a.(k); var33= !cur_int_var }::
            !constraints;

	  if mean_value_subset then begin
	    cur_int_var:= !cur_int_var+1;
   	    constraints:= 
	      Ternary { predicate3= Addition; var31= (List.length !variables)+k; var32= !cur_int_var-1; var33= !cur_int_var }::
	      SubSet { subset_var1= !var_deriv_eval.(k); subset_var2= !cur_int_var }::
	      !constraints
	  end else 
	    constraints:= 
	      Ternary { predicate3= Addition; var31= (List.length !variables)+k; var32= !cur_int_var; var33= !var_deriv_eval.(k) }::!constraints;
	done;
	if trace then begin
	  print_endline "(mean_value_shared";
	  List.iter print_constr !constraints;
	  print_endline "mean_value_shared)";
	end;
	!constraints


        let subtraction_constraints (t1: term_info) (t2: term_info): term_info= 
	  let new_constraints= ref []
	  and term_var= ref (-1)
	  and sample_eval_var= ref (-1)
	  and deriv_eval_var= ref (-1)
	  and deriv_vars= ref (Array.make 0 0) in
	  
	  cur_int_var:= !cur_int_var+1;
		  term_var:= !cur_int_var; (* we need this in any case, because we create mean value constraint recursively *)
	  if original_constraint then begin
	    new_constraints:= Ternary { predicate3= Subtraction; var31= t1.term_var; var32= t2.term_var; var33= !cur_int_var }::!new_constraints;
	  end;
	  if !mean_value then begin
	    cur_int_var:= !cur_int_var+1; (* value at sample point *)
	    sample_eval_var:= !cur_int_var;
	    new_constraints:= Ternary { predicate3= Subtraction; var31= t1.sample_eval_var; var32= t2.sample_eval_var; var33= !cur_int_var }::!new_constraints;
	    
	    cur_int_var:= !cur_int_var+1; (* value at derivative evaluation *)
	    deriv_eval_var:= !cur_int_var;
	    new_constraints:= Ternary { predicate3= Subtraction; var31= t1.deriv_eval_var; var32= t2.deriv_eval_var; var33= !cur_int_var }::!new_constraints;
	    
	    (* propagation of derivatives *)
	    deriv_vars:= Array.make (List.length !variables) 0;
	    for k= 0 to (List.length !variables)-1 do
	      cur_int_var:= !cur_int_var+1; 
	      !deriv_vars.(k)<- !cur_int_var;
	      new_constraints:= Ternary { predicate3= Subtraction; var31= t1.deriv_vars.(k); var32= t2.deriv_vars.(k); var33= !cur_int_var }::!new_constraints;
	    done;
	    
            (* mean_value constraint itself *)
	    if mean_value_recursive then add_mean_value new_constraints !term_var !sample_eval_var !deriv_vars;
	  end;
          { constraints= !new_constraints @ (t1.constraints) @ (t2.constraints); term_var= !term_var; sample_eval_var= !sample_eval_var; deriv_eval_var= !deriv_eval_var; deriv_vars= !deriv_vars; }
	    
	
let error_no_mean_val () =
  print_endline "Mean value constraint not available for function symbol"; exit(1)


    %} 

    %token AND OR IMPLIES EXISTS EXISTS_NOBRANCH FORALL FORALL_NOBRANCH
    %token EQ GE GT LE LT NE SUBSET POSSEMDEF
    %token SIN COS TAN EXP LOG ASIN ACOS ATAN
    %token PLUS MINUS POW
    %token TIMES DIV
    %token SQR
    %token LPAREN RPAREN
    %token END
    %token ZERO
    %token <string> FLOAT
    %token <Interval.interval> INTERVAL
    %token <string> VARIABLE
    %token <int> BOOLVAR
    %token TRUE FALSE
    %token REDUNDANT

    %left AND OR
    %nonassoc REDUNDANT IMPLIES
    %nonassoc EXISTS EXISTS_NOBRANCH FORALL FORALL_NOBRANCH
    %nonassoc SIN COS TAN EXP ASIN ACOS ATAN
    %left PLUS MINUS        /* lowest precedence */
    %left TIMES DIV         /* medium precedence */
    %left UMINUS
    %right SQR, POW   /* highest precedence */
    %token LBRACKET RBRACKET COMMA
    %start expr, box, var_list, single_term             /* the entry points */

    %type <Constr.constr> expr 
    %type <Constr.variable * Constr.constr list> single_term
    %type <Box.box> box
    %type <string list> var_list
    %type <float> signed_float
    %type <int> integer
    %%

    expr:       init lexpr END              { $2 }

    init:                                   { ()  }

  lexpr:      BOOLVAR {
                BoolVar $1
              } 
              | atomic_init atomic                 { 
		  let constraints= ref $2 in
		  if !mean_value then constraints:= !constraints @ (mean_value_shared ());
		  (* in the case of mean value constraint, in addition we have variables representing the sample point *)
                  let temp_var_num= !cur_int_var-free_var_blocks*(List.length !variables)+1 in 
		  let numvars= (free_var_blocks*(List.length !variables)+temp_var_num) in
(*		      let quantvars = Util.int_list (!cur_int_var+1) (numvars-1) in *)
                  let atomic= 
		    let branches= !constraints in
		    Atomic { obound= Box.full_box temp_var_num; ibound= Box.full_box temp_var_num; 
			     aconstr= Conjunction { branches= branches; bloated= true; next_branch= 0; changed= Util.int_list 0 ((List.length $2)-1); wakeup_table= Constr.make_wakeup_table branches numvars; maxconj_pos= false; maxconj_neg= false };
			   } in
	          if !mean_value then 
		    (* introduce variables for sample_point evaluation *)
		    UnivQuantBound { var_trans= Array.init (List.length !variables ) (function k->k); qbsubconstr= atomic }
		  else atomic
                }
              | EXISTS var_list_env box lexpr { 
                            		  if not ((List.length $2)=(Box.dim $3)) then raise Parsing.Parse_error;
		                          variables:= Util.drop_last !variables (List.length $2);
                                          let cur_int_var= (List.length !variables-1) in
		                          let quantvars= Util.int_list (cur_int_var+1) (cur_int_var+(Box.dim $3)) in
					  ExistQuant { qbound= $3; subconstr= $4; outer_changed= quantvars; inner_changed= quantvars; qnext_branch= 0 } 
                                       }
              | EXISTS_NOBRANCH var_list_env box lexpr { 
                            		  if not ((List.length $2)=(Box.dim $3)) then raise Parsing.Parse_error;
		                          variables:= Util.drop_last !variables (List.length $2);
                                          let cur_int_var= (List.length !variables-1) in
		                          let quantvars= Util.int_list (cur_int_var+1) (cur_int_var+(Box.dim $3)) in
					  ExistQuant { qbound= $3; subconstr= $4; outer_changed= quantvars; inner_changed= quantvars; qnext_branch= -1 } 
                                       } 
              | FORALL var_list_env box lexpr { 
		                         if not ((List.length $2)=(Box.dim $3)) then raise Parsing.Parse_error;
                                         variables:= Util.drop_last !variables (List.length $2);
                                         let cur_int_var= (List.length !variables-1) in
		                         let quantvars= Util.int_list (cur_int_var+1) (cur_int_var+(Box.dim $3)) in
		                         UnivQuant { qbound= $3; subconstr= $4; outer_changed= quantvars; inner_changed= quantvars; qnext_branch= 0 } 
                                       }
              | FORALL_NOBRANCH var_list_env box lexpr { 
		                         if not ((List.length $2)=(Box.dim $3)) then raise Parsing.Parse_error;
                                         variables:= Util.drop_last !variables (List.length $2);
                                         let cur_int_var= (List.length !variables-1) in
		                         let quantvars= Util.int_list (cur_int_var+1) (cur_int_var+(Box.dim $3)) in
		                         UnivQuant { qbound= $3; subconstr= $4; outer_changed= quantvars; inner_changed= quantvars; qnext_branch= -1 } 
                                       } 
              | lexpr AND lexpr        { 
		  let br1 = 
		    match $1 with
		      Conjunction c -> c.branches |
		      _ -> [ $1 ] 
		  and br2 = 
		    match $3 with
		      Conjunction c -> c.branches |
		      _ -> [ $3 ] in
		  let br= br1 @ br2 in 
		  Conjunction (build_connective br (List.length !variables)) (* re-builds wake-up tables, which is inefficient *)
                }
              | lexpr OR lexpr         { 
		  let br1 = 
		    match $1 with
		      Disjunction c -> c.branches |
		      _ -> [ $1 ] 
		  and br2 = 
		    match $3 with
		      Disjunction c -> c.branches |
		      _ -> [ $3 ] in
		  let br= br1 @ br2 in
		  Disjunction (build_connective br (List.length !variables)) (* re-builds wake-up tables, which is inefficient *)
}
              | lexpr IMPLIES lexpr         { 
		  let br1 = 
		    match $1 with
		      Conjunction c -> c.branches |
		      _ -> [ $1 ] 
		  and br2 = 
		    match $3 with
		      Disjunction c -> c.branches |
		      _ -> [ $3 ] in
		  let br= (List.map Constr.neg br1) @ br2 in
		  Disjunction (build_connective br (List.length !variables)) (* re-builds wake-up tables, which is inefficient *)
}
              | lexpr REDUNDANT lexpr { 
		                         Redundant { original_constr = $1; redundant_constr = $3 }
                                       }
              |	LBRACKET lexpr RBRACKET    { $2 }
    ;

		atomic_init:             { 
   		  cur_int_var:= (List.length !variables-1);
		  if !mean_value then begin
		    let varnum= (List.length !variables) in
                    (* variables introduced for sample point *)    
		    cur_int_var:= !cur_int_var+varnum;

		    (* variable vector y, at which derivatives are evaluated *)
		    cur_int_var:= !cur_int_var+varnum;
		    var_deriv_eval:= Array.init varnum (function v -> !cur_int_var-v);

		    (* derivative of each variable wrt. to each other variable *)
		    cur_int_var:= !cur_int_var+varnum*varnum;
		    var_deriv:= Array.init varnum (function v -> Array.init varnum (function d -> !cur_int_var-v*varnum-d));
		    
		    (* variable vector for x-a *)
		    cur_int_var:= !cur_int_var+varnum;
		    var_x_minus_a:= Array.init varnum (function v -> !cur_int_var-v);
		  end
                } 



   atomic:      TRUE                    { [BoolIntv { Bool_intv.lo= true; Bool_intv.hi= true }] } 
	      | FALSE                   { [BoolIntv { Bool_intv.lo= false; Bool_intv.hi= false }] } 
              | term EQ ZERO            { 
		let constraints= ref $1.constraints in
		if !mean_value && not mean_value_recursive then
		  add_mean_value constraints $1.term_var $1.sample_eval_var $1.deriv_vars;
		constraints:= Unary { predicate1= Equal; var1= $1.term_var }::!constraints;
		!constraints
                }
              | term NE ZERO            { 
		let constraints= ref $1.constraints in
		if !mean_value && not mean_value_recursive then
		  add_mean_value constraints $1.term_var $1.sample_eval_var $1.deriv_vars;
		constraints:= Unary { predicate1= NotEqual; var1= $1.term_var }::!constraints;
		!constraints
                } |
              | term GT ZERO            { 
		let constraints= ref $1.constraints in
		if !mean_value && not mean_value_recursive then
		  add_mean_value constraints $1.term_var $1.sample_eval_var $1.deriv_vars;
		constraints:= Unary { predicate1= Greater; var1= $1.term_var }::!constraints;
		!constraints
                } |
              | term GE ZERO            { 
		let constraints= ref $1.constraints in
		if !mean_value && not mean_value_recursive then
		  add_mean_value constraints $1.term_var $1.sample_eval_var $1.deriv_vars;
		constraints:= Unary { predicate1= GreaterEqual; var1= $1.term_var }::!constraints;
		!constraints
                } |
              | term LT ZERO            { 
		let constraints= ref $1.constraints in
		if !mean_value && not mean_value_recursive then
		  add_mean_value constraints $1.term_var $1.sample_eval_var $1.deriv_vars;
		constraints:= Unary { predicate1= Less; var1= $1.term_var }::!constraints;
		!constraints
                } |
              | term LE ZERO            { 
		let constraints= ref $1.constraints in
		if !mean_value && not mean_value_recursive then
		  add_mean_value constraints $1.term_var $1.sample_eval_var $1.deriv_vars;
		constraints:= Unary { predicate1= LessEqual; var1= $1.term_var }::!constraints;
		!constraints
                } |
              |	term EQ term            { 
		  let term_info= subtraction_constraints $1 $3 in 
		  let constraints= ref term_info.constraints in
		  if !mean_value && not mean_value_recursive then
		    add_mean_value constraints term_info.term_var term_info.sample_eval_var term_info.deriv_vars;
                  Unary { predicate1= Equal; var1= term_info.term_var }::!constraints;
                }
              |	term NE term            { 
		  let term_info= subtraction_constraints $1 $3 in 
		  let constraints= ref term_info.constraints in
		  if !mean_value && not mean_value_recursive then
		    add_mean_value constraints term_info.term_var term_info.sample_eval_var term_info.deriv_vars;
                  Unary { predicate1= NotEqual; var1= term_info.term_var }::!constraints;
                }
              |	term GT term            { 
		  let term_info= subtraction_constraints $1 $3 in 
		  let constraints= ref term_info.constraints in
		  if !mean_value && not mean_value_recursive then
		    add_mean_value constraints term_info.term_var term_info.sample_eval_var term_info.deriv_vars;
                  Unary { predicate1= Greater; var1= term_info.term_var }::!constraints;
                }
              |	term GE term            { 
		  let term_info= subtraction_constraints $1 $3 in 
		  let constraints= ref term_info.constraints in
		  if !mean_value && not mean_value_recursive then
		    add_mean_value constraints term_info.term_var term_info.sample_eval_var term_info.deriv_vars;
                  Unary { predicate1= GreaterEqual; var1= term_info.term_var }::!constraints;
                }
              |	term LT term            { 
		  let term_info= subtraction_constraints $1 $3 in 
		  let constraints= ref term_info.constraints in
		  if !mean_value && not mean_value_recursive then
		    add_mean_value constraints term_info.term_var term_info.sample_eval_var term_info.deriv_vars;
                  Unary { predicate1= Less; var1= term_info.term_var }::!constraints;
                }
              |	term LE term            { 
		  let term_info= subtraction_constraints $1 $3 in 
		  let constraints= ref term_info.constraints in
		  if !mean_value && not mean_value_recursive then
		    add_mean_value constraints term_info.term_var term_info.sample_eval_var term_info.deriv_vars;
                  Unary { predicate1= LessEqual; var1= term_info.term_var }::!constraints;
                }
              |	term SUBSET term            { 
                  SubSet { subset_var1= $1.term_var; subset_var2= $3.term_var }::(($1.constraints) @ ($3.constraints)); 
                }
              |	POSSEMDEF LPAREN term_matrix RPAREN       { (PosSemDef (fst $3))::(snd $3) } 




      ;

    term_matrix:   LBRACKET term_list_seq RBRACKET { (Array.of_list (fst $2), (snd $2)) }
		  
    term_list_seq: term_list {  ([fst $1], snd $1)  }
	      | term_list COMMA term_list_seq { ((fst $1)::(fst $3), (snd $1) @ (snd $3))  }
      
    term_list:  LBRACKET term_seq RBRACKET { (Array.of_list (fst $2), snd $2) }

   term_seq:    term  { ([ $1.term_var ], $1.constraints)  }
	      | term COMMA term_seq { (($1.term_var)::(fst $3), ($1.constraints) @ (snd $3)) }
    

    /* type: variable number */
    term:       constant {
	          let new_constraints= ref [] 
		  and term_var= ref (-1)
		  and sample_eval_var= ref (-1)
		  and deriv_eval_var= ref (-1)
	          and deriv_vars= ref (Array.make 0 0) in
                  cur_int_var:= !cur_int_var+1; 
		  term_var:= !cur_int_var;
		  new_constraints:= IsElement { var= !cur_int_var ; interval= $1  }::!new_constraints;
		  if !mean_value then begin
		    (* cur_int_var:= !cur_int_var+1;*) (* value at sample point and derivative evluation *)
		    sample_eval_var:= !cur_int_var; deriv_eval_var:= !cur_int_var;
(*		    new_constraints:= IsElement { var= !cur_int_var ; interval= $1  }::!new_constraints; *)

		    let deriv k= begin
		      cur_int_var:= !cur_int_var+1; 
		      new_constraints:= IsElement { var= !cur_int_var ; interval= { lo= 0.0; hi= 0.0  } }::!new_constraints;
		      !cur_int_var
		    end in
		    deriv_vars:= Array.init (List.length !variables) deriv;
		  end;
                  { constraints= !new_constraints; term_var= !term_var; sample_eval_var= !sample_eval_var; deriv_eval_var= !deriv_eval_var; deriv_vars= !deriv_vars; }
                }
              | LPAREN term RPAREN      { $2 } 
               | term POW integer        { 
		  let constraints = ref []
		  and sample_eval_var= ref (-1) 
		  and deriv_eval_var= ref (-1) 
	          and deriv_vars= ref (Array.make 0 0) in
                  cur_int_var:= !cur_int_var+1; (* the variable representing the current term tree *)
		  let term_var= !cur_int_var in
		  if original_constraint then begin
		    constraints:= Power  {var01= $1.term_var; n= $3; var02 = !cur_int_var }::!constraints;
		  end;
		  if !mean_value then begin
		    cur_int_var:= !cur_int_var+1; (* value at sample point *)
		    sample_eval_var:= !cur_int_var;
		    constraints:= Power  {var01= $1.sample_eval_var; n= $3; var02 = !cur_int_var }::!constraints;

		    cur_int_var:= !cur_int_var+1; (* value at derivative evaluation *)
		    deriv_eval_var:= !cur_int_var;
		    constraints:= Power  {var01= $1.deriv_eval_var; n= $3; var02 = !cur_int_var }::!constraints;

		    (* propagation of derivatives *)
		    deriv_vars:= Array.make (List.length !variables) 0;
		    cur_int_var:= !cur_int_var+1; 
		    let constvar= !cur_int_var in
		    constraints:= IsElement { var= constvar; interval= { lo= float_of_int $3; hi= float_of_int $3 } }::!constraints;
		    for k= 0 to (List.length !variables)-1 do
		      cur_int_var:= !cur_int_var+3; 
		      !deriv_vars.(k)<- !cur_int_var;
		      constraints:= 
			Power { var01= $1.deriv_eval_var; n= $3-1; var02= !cur_int_var-2 }::
			Ternary { predicate3= Multiplication; var31= !cur_int_var-2; var32= $1.deriv_vars.(k); var33= !cur_int_var-1 }::
			Ternary { predicate3= Multiplication; var31= constvar; var32= !cur_int_var-1; var33= !cur_int_var }::
			!constraints; 
		    done;

		    if trace then begin
		      print_endline "(term POW integer";
		      List.iter print_constr !constraints;
		      print_endline "term POW integer)";
		    end;

                    (* mean_value constraint itself *)
		    if mean_value_recursive then add_mean_value constraints term_var !sample_eval_var !deriv_vars;
		  end;
		{ constraints= !constraints @ $1.constraints;
		  term_var= term_var;
		  sample_eval_var= !sample_eval_var;
		  deriv_eval_var= !deriv_eval_var; 
		  deriv_vars= !deriv_vars;
		}
      }
              | term PLUS term         { 
		  let new_constraints= ref []
		  and term_var= ref (-1)
		  and sample_eval_var= ref (-1)
		  and deriv_eval_var= ref (-1)
		  and deriv_vars= ref (Array.make 0 0) in

		  cur_int_var:= !cur_int_var+1;
		  term_var:= !cur_int_var; (* we need this in any case, because we create mean value constraint recursively *)
		  if original_constraint then begin
		    new_constraints:= Ternary { predicate3= Addition; var31= $1.term_var; var32= $3.term_var; var33= !cur_int_var }::!new_constraints;
		  end;
		  if !mean_value then begin
		    cur_int_var:= !cur_int_var+1; (* value at sample point *)
		    sample_eval_var:= !cur_int_var;
		    new_constraints:= Ternary { predicate3= Addition; var31= $1.sample_eval_var; var32= $3.sample_eval_var; var33= !cur_int_var }::!new_constraints;

		    cur_int_var:= !cur_int_var+1; (* value at derivative evaluation *)
		    deriv_eval_var:= !cur_int_var;
		    new_constraints:= Ternary { predicate3= Addition; var31= $1.deriv_eval_var; var32= $3.deriv_eval_var; var33= !cur_int_var }::!new_constraints;

		    (* propagation of derivatives *)
		    deriv_vars:= Array.make (List.length !variables) 0;
		    for k= 0 to (List.length !variables)-1 do
		      cur_int_var:= !cur_int_var+1; 
		      !deriv_vars.(k)<- !cur_int_var;
		      new_constraints:= Ternary { predicate3= Addition; var31= $1.deriv_vars.(k); var32= $3.deriv_vars.(k); var33= !cur_int_var }::!new_constraints;
		    done;
		    if trace then begin
		      print_endline "(term PLUS term";
		      List.iter print_constr !new_constraints;
		      print_endline "term PLUS term)";
		    end;

                    (* mean_value constraint itself *)
		    if mean_value_recursive then add_mean_value new_constraints !term_var !sample_eval_var !deriv_vars;
		  end;
                  { constraints= !new_constraints @ ($1.constraints) @ ($3.constraints); term_var= !term_var; sample_eval_var= !sample_eval_var; deriv_eval_var= !deriv_eval_var; deriv_vars= !deriv_vars; }
                } 
              | term MINUS term         { 
		subtraction_constraints $1 $3
                } 
              | term TIMES term         { 
		  let new_constraints= ref []
		  and term_var= ref (-1)
                  and deriv_eval_var= ref (-1)
		  and sample_eval_var= ref (-1)
		  and deriv_vars= ref (Array.make 0 0) in

		  cur_int_var:= !cur_int_var+1;
		  term_var:= !cur_int_var; (* we need this in any case, because we create mean value constraint recursively *)
		  if original_constraint then begin
		    new_constraints:= Ternary { predicate3= Multiplication; var31= $1.term_var; var32= $3.term_var; var33= !cur_int_var }::!new_constraints;
		  end;
		  if !mean_value then begin
		    cur_int_var:= !cur_int_var+1; (* value at sample point *)
		    sample_eval_var:= !cur_int_var;
		    new_constraints:= Ternary { predicate3= Multiplication; var31= $1.sample_eval_var; var32= $3.sample_eval_var; var33= !cur_int_var }::!new_constraints;

		    cur_int_var:= !cur_int_var+1; (* value at derivative evaluation *)
		    deriv_eval_var:= !cur_int_var;
		    new_constraints:= Ternary { predicate3= Multiplication; var31= $1.deriv_eval_var; var32= $3.deriv_eval_var; var33= !cur_int_var }::!new_constraints;


		    (* propagation of derivatives *)
		    deriv_vars:= Array.make (List.length !variables) 0;
		    for k= 0 to (List.length !variables)-1 do
		      cur_int_var:= !cur_int_var+1;
		      new_constraints:= Ternary { predicate3= Multiplication; var31= $1.deriv_vars.(k); var32= $3.deriv_eval_var; var33= !cur_int_var }::!new_constraints;
		      
		      cur_int_var:= !cur_int_var+1;
		      new_constraints:= Ternary { predicate3= Multiplication; var31= $1.deriv_eval_var; var32= $3.deriv_vars.(k); var33= !cur_int_var }::!new_constraints;
		      
		      cur_int_var:= !cur_int_var+1; 
		      !deriv_vars.(k)<- !cur_int_var;
		      new_constraints:= Ternary { predicate3= Addition; var31= !cur_int_var-2; var32= !cur_int_var-1; var33= !cur_int_var }::!new_constraints;
		    done;
		    if trace then begin
		      print_endline "(term TIMES term";
		      List.iter print_constr !new_constraints;
		      print_endline "term TIMES term)";
		    end;
                    (* mean_value constraint itself *)
		    if mean_value_recursive then add_mean_value new_constraints !term_var !sample_eval_var !deriv_vars;
		  end;
                  { constraints= !new_constraints @ ($1.constraints) @ ($3.constraints); term_var= !term_var; sample_eval_var= !sample_eval_var; deriv_eval_var= !deriv_eval_var; deriv_vars= !deriv_vars; }
                }
              |	MINUS term %prec UMINUS {
		  let constraints = ref $2.constraints
		  and sample_eval_var= ref (-1) 
		  and deriv_eval_var= ref (-1) 
	          and deriv_vars= ref (Array.make 0 0) in

                  cur_int_var:= !cur_int_var+1; (* the variable representing the current term tree *)
		  let term_var= !cur_int_var in
		  if original_constraint then begin
		    constraints:= Binary { predicate2= UMinus; var21= $2.term_var; var22= !cur_int_var }::!constraints;
		  end;
		  if !mean_value then begin
		    cur_int_var:= !cur_int_var+1; (* value at sample point *)
		    sample_eval_var:= !cur_int_var;
		    constraints:= Binary { predicate2= UMinus; var21= $2.sample_eval_var; var22= !cur_int_var }::!constraints;

		    cur_int_var:= !cur_int_var+1; (* value at derivative evaluation *)
		    deriv_eval_var:= !cur_int_var;
		    constraints:= Binary { predicate2= UMinus; var21= $2.deriv_eval_var; var22= !cur_int_var }::!constraints;

		    (* propagation of derivatives *)
		    deriv_vars:= Array.make (List.length !variables) 0;
		    for k= 0 to (List.length !variables)-1 do
		      cur_int_var:= !cur_int_var+1; 
		      !deriv_vars.(k)<- !cur_int_var;
		      constraints:= 
			Binary { predicate2= UMinus; var21= $2.deriv_vars.(k); var22= !cur_int_var }::
			!constraints; 
		    done;

                    (* mean_value constraint itself *)
		    if mean_value_recursive then add_mean_value constraints term_var !sample_eval_var !deriv_vars;
		  end;
		{ constraints= !constraints;
		  term_var= term_var;
		  sample_eval_var= !sample_eval_var;
		  deriv_eval_var= !deriv_eval_var; 
		  deriv_vars= !deriv_vars;
                }
      }
              | SIN LPAREN term RPAREN                { 
		  let constraints = ref $3.constraints
		  and sample_eval_var= ref (-1) 
		  and deriv_eval_var= ref (-1) 
	          and deriv_vars= ref (Array.make 0 0) in

		  cur_int_var:= !cur_int_var+1; 
		  let term_var= !cur_int_var in
		  if original_constraint then begin
		    constraints:= Binary { predicate2= Sin; var21= $3.term_var; var22= term_var }::!constraints;
		  end;
		  if !mean_value then begin
		    cur_int_var:= !cur_int_var+1; (* value at sample point *)
		    sample_eval_var:= !cur_int_var;
		    constraints:= Binary { predicate2= Sin; var21= $3.sample_eval_var; var22= !sample_eval_var }::!constraints;

		    cur_int_var:= !cur_int_var+1; (* value at derivative evaluation *)
		    deriv_eval_var:= !cur_int_var;
		    constraints:= Binary { predicate2= Sin; var21= $3.deriv_eval_var; var22= !deriv_eval_var }::!constraints;

		    (* propagation of derivatives *)
		    deriv_vars:= Array.make (List.length !variables) 0;
		    for k= 0 to (List.length !variables)-1 do
		      cur_int_var:= !cur_int_var+2; 
		      !deriv_vars.(k)<- !cur_int_var;
		      constraints:= 
			Binary { predicate2= Cos; var21= $3.deriv_eval_var; var22= !cur_int_var-1 }::
			Ternary { predicate3= Multiplication; var31= $3.deriv_vars.(k); var32= !cur_int_var-1; var33= !cur_int_var }::
			!constraints; 
		    done;

                    (* mean_value constraint itself *)
		    if mean_value_recursive then add_mean_value constraints term_var !sample_eval_var !deriv_vars;
		  end;
		  
		  { constraints= !constraints;
		    term_var= term_var;
		    sample_eval_var= !sample_eval_var;
		    deriv_eval_var= !deriv_eval_var;
		    deriv_vars= !deriv_vars
		  }
                }
              | COS LPAREN term RPAREN                { 
		  let constraints = ref $3.constraints
		  and sample_eval_var= ref (-1) 
		  and deriv_eval_var= ref (-1) 
	          and deriv_vars= ref (Array.make 0 0) in

		  cur_int_var:= !cur_int_var+1; 
		  let term_var= !cur_int_var in
		  if original_constraint then begin
		    constraints:= Binary { predicate2= Cos; var21= $3.term_var; var22= term_var }::!constraints;
		  end;
		  if !mean_value then begin
		    cur_int_var:= !cur_int_var+1; (* value at sample point *)
		    sample_eval_var:= !cur_int_var;
		    constraints:= Binary { predicate2= Cos; var21= $3.sample_eval_var; var22= !sample_eval_var }::!constraints;

		    cur_int_var:= !cur_int_var+1; (* value at derivative evaluation *)
		    deriv_eval_var:= !cur_int_var;
		    constraints:= Binary { predicate2= Cos; var21= $3.deriv_eval_var; var22= !deriv_eval_var }::!constraints;

		    (* propagation of derivatives *)
		    deriv_vars:= Array.make (List.length !variables) 0;
		    for k= 0 to (List.length !variables)-1 do
		      cur_int_var:= !cur_int_var+3; 
		      !deriv_vars.(k)<- !cur_int_var;
		      constraints:= 
			Binary { predicate2= Sin; var21= $3.deriv_eval_var; var22= !cur_int_var-2 }::
			Ternary { predicate3= Multiplication; var31= $3.deriv_vars.(k); var32= !cur_int_var-2; var33= !cur_int_var-1 }::
			Binary{ predicate2= UMinus; var21= !cur_int_var-1; var22= !cur_int_var }::
			!constraints; 
		    done;

                    (* mean_value constraint itself *)
		    if mean_value_recursive then add_mean_value constraints term_var !sample_eval_var !deriv_vars;
		  end;
		  
		  { constraints= !constraints;
		    term_var= term_var;
		    sample_eval_var= !sample_eval_var;
		    deriv_eval_var= !deriv_eval_var;
		    deriv_vars= !deriv_vars
		  }
                }
              | TAN LPAREN term RPAREN                { 
		  let constraints = ref $3.constraints
		  and sample_eval_var= ref (-1) 
		  and deriv_eval_var= ref (-1) 
	          and deriv_vars= ref (Array.make 0 0) in

		  cur_int_var:= !cur_int_var+1; 
		  let term_var= !cur_int_var in
		  if original_constraint then begin
		    constraints:= Binary { predicate2= Tan; var21= $3.term_var; var22= term_var }::!constraints;
		  end;
		  if !mean_value then error_no_mean_val ();
		  { constraints= !constraints;
		    term_var= term_var;
		    sample_eval_var= !sample_eval_var;
		    deriv_eval_var= !deriv_eval_var;
		    deriv_vars= !deriv_vars
		  }
                }
              | ASIN LPAREN term RPAREN                { 
		  let sample_eval_var= ref (-1) in
		  if !mean_value then error_no_mean_val ();
		   cur_int_var:= !cur_int_var+1;                           
		  { constraints= Binary { predicate2= ArcSin; var21= $3.term_var; var22= !cur_int_var }::($3.constraints);
		    term_var= !cur_int_var;
		    sample_eval_var= !sample_eval_var;
		    deriv_eval_var= -1; 
		    deriv_vars= Array.make 0 0 
		  }
                }
              | ACOS LPAREN term RPAREN                { 
		  let sample_eval_var= ref (-1) in
		  if !mean_value then error_no_mean_val ();
		  cur_int_var:= !cur_int_var+1; 
		  { constraints= Binary { predicate2= ArcCos; var21= $3.term_var; var22= !cur_int_var }::($3.constraints);
		    term_var= !cur_int_var;
		    sample_eval_var= !sample_eval_var;
		    deriv_eval_var= -1; 
		    deriv_vars= Array.make 0 0
		  }
                }
              | ATAN LPAREN term RPAREN                { 
		  let sample_eval_var= ref (-1) in
		  if !mean_value then error_no_mean_val ();
		   cur_int_var:= !cur_int_var+1;                           
		  { constraints= Binary { predicate2= ArcTan; var21= $3.term_var; var22= !cur_int_var }::($3.constraints);
		    term_var= !cur_int_var;
		    sample_eval_var= !sample_eval_var;
		    deriv_eval_var= -1; 
		    deriv_vars= Array.make 0 0
		  }
                }
              | EXP LPAREN term RPAREN                { 
		  let constraints = ref $3.constraints
		  and sample_eval_var= ref (-1) 
		  and deriv_eval_var= ref (-1) 
	          and deriv_vars= ref (Array.make 0 0) in

		  cur_int_var:= !cur_int_var+1; 
		  let term_var= !cur_int_var in
		  if original_constraint then begin
		    constraints:= Binary { predicate2= Exp; var21= $3.term_var; var22= term_var }::!constraints;
		  end;
		  if !mean_value then begin
		    cur_int_var:= !cur_int_var+1; (* value at sample point *)
		    sample_eval_var:= !cur_int_var;
		    constraints:= Binary { predicate2= Exp; var21= $3.sample_eval_var; var22= !sample_eval_var }::!constraints;

		    cur_int_var:= !cur_int_var+1; (* value at derivative evaluation *)
		    deriv_eval_var:= !cur_int_var;
		    constraints:= Binary { predicate2= Exp; var21= $3.deriv_eval_var; var22= !deriv_eval_var }::!constraints;

		    (* propagation of derivatives *)
		    deriv_vars:= Array.make (List.length !variables) 0;
		    for k= 0 to (List.length !variables)-1 do
		      cur_int_var:= !cur_int_var+2; 
		      !deriv_vars.(k)<- !cur_int_var;
		      constraints:= 
			Binary { predicate2= Exp; var21= $3.deriv_eval_var; var22= !cur_int_var-1 }::
			Ternary { predicate3= Multiplication; var31= $3.deriv_vars.(k); var32= !cur_int_var-1; var33= !cur_int_var }::
			!constraints; 
		    done;

                    (* mean_value constraint itself *)
		    if mean_value_recursive then add_mean_value constraints term_var !sample_eval_var !deriv_vars;
		  end;
		  
		  { constraints= !constraints;
		    term_var= term_var;
		    sample_eval_var= !sample_eval_var;
		    deriv_eval_var= !deriv_eval_var;
		    deriv_vars= !deriv_vars
		  }
                }
              | LOG LPAREN term RPAREN                { 
		  let sample_eval_var= ref (-1) in
		  if !mean_value then error_no_mean_val ();
		  cur_int_var:= !cur_int_var+1;        
		{ constraints= Binary { predicate2= Log; var21= $3.term_var; var22= !cur_int_var }::($3.constraints);
		  term_var= !cur_int_var;
		  sample_eval_var= !sample_eval_var;
		  deriv_eval_var= -1; 
		  deriv_vars= Array.make 0 0
		}
}
              | term SQR                { 
		  let constraints = ref []
		  and sample_eval_var= ref (-1) 
		  and deriv_eval_var= ref (-1) 
	          and deriv_vars= ref (Array.make 0 0) in
                  cur_int_var:= !cur_int_var+1; (* the variable representing the current term tree *)
		  let term_var= !cur_int_var in
		  if original_constraint then begin
		    constraints:= Binary { predicate2= Square; var21= $1.term_var; var22= !cur_int_var }::!constraints;
		  end;
		  if !mean_value then begin
		    cur_int_var:= !cur_int_var+1; (* value at sample point *)
		    sample_eval_var:= !cur_int_var;
		    constraints:= Binary { predicate2= Square; var21= $1.sample_eval_var; var22= !cur_int_var }::!constraints;

		    cur_int_var:= !cur_int_var+1; (* value at derivative evaluation *)
		    deriv_eval_var:= !cur_int_var;
		    constraints:= Binary { predicate2= Square; var21= $1.deriv_eval_var; var22= !cur_int_var }::!constraints;

		    (* propagation of derivatives *)
		    deriv_vars:= Array.make (List.length !variables) 0;
		    for k= 0 to (List.length !variables)-1 do
		      cur_int_var:= !cur_int_var+2; 
		      !deriv_vars.(k)<- !cur_int_var;
		      constraints:= (* we add the same variable instead of multiplying by 2 *)
			Ternary { predicate3= Multiplication; var31= $1.deriv_eval_var; var32= $1.deriv_vars.(k); var33= !cur_int_var-1 }::
			Ternary { predicate3= Addition; var31= !cur_int_var-1; var32= !cur_int_var-1; var33= !cur_int_var }::
			!constraints; 
		    done;

		    if trace then begin
		      print_endline "(term SQR";
		      List.iter print_constr !constraints;
		      print_endline "term SQR)";
		    end;

                    (* mean_value constraint itself *)
		    if mean_value_recursive then add_mean_value constraints term_var !sample_eval_var !deriv_vars;
		  end;
		{ constraints= !constraints @ $1.constraints;
		  term_var= term_var;
		  sample_eval_var= !sample_eval_var;
		  deriv_eval_var= !deriv_eval_var; 
		  deriv_vars= !deriv_vars;
		}
      }
              | VARIABLE               {  let pos = Util.pos !Parser_intern.variables $1 in 
                                          if pos<0 then begin
                                            print_endline ("Variable " ^ $1 ^ " not found"); exit(1)
                                          end else 
					    let deriv_vars= 
					      if !mean_value then !var_deriv.(pos)
					      else Array.make 0 0 in
					    { constraints= 
					        if !mean_value then 
						  let init k= 
						    IsElement { var= deriv_vars.(k); 
								interval= if k=pos then { lo= 1.0; hi= 1.0 } else { lo= 0.0; hi= 0.0 } } in
						  Array.to_list (Array.init (List.length !variables) init)
   					        else []; 
					      term_var= pos; 
					      sample_eval_var= 
					        if !mean_value then pos+(List.length !variables) else (-1); 
					      deriv_eval_var= 
					        if !mean_value then !var_deriv_eval.(pos) else (-1); 
					      deriv_vars= deriv_vars }
                                        }

    constant:   FLOAT         	{  (* for some strange reason, float_of_string cannot parse "neg_infinity" *)
				   if $1="neg_infinity" then { Interval.lo=neg_infinity ; Interval.hi=neg_infinity }
				   else try
					let f = float_of_int( int_of_string($1) )
					in { Interval.lo=f ; Interval.hi=f }
				   with
					Failure("int_of_string") -> 
					  let f = float_of_string($1)
					  in { Interval.lo= prev_fp(f); Interval.hi= next_fp(f) }
				} 
	      | ZERO              { { Interval.lo= 0.0; Interval.hi= 0.0 } } 
              |	INTERVAL          { $1 } 
				       
    box:        LBRACKET RBRACKET              { Box.zero_dim () }
	      |	LBRACKET interval_seq RBRACKET { Box.make (Array.of_list $2) }

    interval_seq: interval                       { [ $1 ] }
              | interval COMMA interval_seq    { $1 :: $3 }

/*    interval:   LBRACKET signed_float COMMA signed_float RBRACKET { { lo= $2; hi= $4 } } */
    interval:   INTERVAL                         { $1 }

    integer:   signed_float	               { if not (floor($1)=$1) then begin
                                                   print_endline "Only integer exponents allowed!";
                                                   exit(1);
                                                 end;
					         truncate($1) }

    signed_float: ZERO                         { 0. } 
                 |  float                       { $1 }
               | MINUS float                  { -. $2 }

    float:       FLOAT                          { float_of_string($1) }
/*              | MINUS FLOAT                    { -. $2 } */

    var_list_env: var_list                     { variables:= !variables @ $1; 
						 $1 }
                  
    var_list:   LBRACKET RBRACKET              { [] } 
              |	LBRACKET var_seq RBRACKET      { $2 }
   
    var_seq:    var                            { [ $1 ] }
              | var COMMA var_seq              { $1::$3 }

    var:        VARIABLE                       { $1 }

    single_term: init term END                      { ($2.term_var, $2.constraints) }             
   

