(* interface to semidefinite programming *)
open Interval

type 'a matrix = 'a array array 

type sdp_result =
    SDPSuccess |
    SDPUnbounded |
    SDPNotFeasible |
    SDPNoSuccess 

external solver: unit -> string = "sdp_solver_"

external sdpx: interval matrix  -> float matrix -> sdp_result = "sdpx_"

let call_counter= ref 0

let num_calls ()=
  !call_counter

let sdp (m: interval matrix) (v: int * int) (max: bool): (sdp_result * float) =
  let debug= false in
  let array_make i = Array.make (Array.length m) 0.0 in
  let f= Array.init (Array.length m) array_make in
  let maxvar= Pervasives.max (fst v) (snd v) 
  and minvar= min (fst v) (snd v) in
(*  let f = Array.make_matrix (Array.length m) (Array.length m) 0.0 in *)
  begin
    call_counter:= !call_counter+1;
    if debug then begin print_string "PSD-in"; print_int minvar; print_int maxvar; print_newline (); flush stdout; end;
    f.(maxvar).(minvar) <- if max then 1.0 else -1.0;
    let success= sdpx m f in begin
      if debug then begin print_endline "PSD-out"; print_float f.(maxvar).(minvar); flush stdout; end;
      (success, f.(maxvar).(minvar)) 
    end;
  end

