let fold_map (f: 'a -> 'b) (g: 'b -> 'b -> 'b) (zero: 'b) (list: 'a list) =
  let fold_f a b = g a (f b)
  in (List.fold_left fold_f zero list: 'b)

(* return a pair s.t. 
       first element is maximal and 
       second element is list without first element
*)
let rec get_max (f: 'a -> float) (l: 'a list): ('a * 'a list) =
  let rec get_max_val (f: 'a -> float) (l: 'a list): ('a * 'a list * float ) = (* avoids recomputation *)
    match l with
      [ a ] -> (a, [], f a) |
      (first :: rest) -> 
	let (rest_max, rest_l, rest_val) = get_max_val f rest
	and first_val = f first
	in if first_val>rest_val then (first, rest_max :: rest_l, first_val)
	else (rest_max, first :: rest_l, rest_val) |
	_ -> assert(false) in
  let (max, rest, max_val)= get_max_val f l in
  (max, rest)
	    
let rec max_el (f: 'a -> float) (l: 'a list): int = 
  match l with
  [ a ] -> 0 |
  (first :: rest) -> 
    let (rest_max, rest_l) = get_max f rest
    in if (f first)>(f rest_max) then 0
       else (max_el f rest)+1 |
  _ -> assert(false)

let rec disjoint (a: 'a list) (b: 'b list): bool = 
  match a with
  [] -> true |
  first :: rest -> 
    if List.mem first b then false
    else disjoint rest b

let rec intersection (a: 'a list) (b: 'a list): 'list =
  match a with 
  [] -> [] |
  first :: rest ->
    let rint = intersection rest b in
    if List.mem first b then first::rint 
    else rint

let rec prod (comb: 'a -> 'b -> 'c) (l1: 'a list) (l2: 'b list) =
  match l1 with
  [] -> [] |
  first::rest -> (List.append (prod comb rest l2: 'c list) (List.map (comb first) l2): 'c list)

let rec int_list (from: int) (to_: int): int list =
  if from>to_ then []
  else from::(int_list (from+1) to_)

let rec pos (l: 'a list) (x:'a): int =
  match l with
    [] -> -1 |
    first::rest -> 
      if first=x then 0
      else 
	let r= (pos rest x)
	in if r=(-1) then -1 else r+1

(* inefficient implementation! *)
let rec drop_last (l: 'a list) (n: int): 'a list =
  if (List.length l)<=n then
    []
  else 
    (List.hd l)::(drop_last (List.tl l) n)

let rec unique (l: 'a list): bool =
  match l with
    [] -> true |
    first::rest -> 
      if List.mem first rest then false
      else unique rest

let compose f g x =
  f (g x)

let rec count (p: 'a -> bool) (l: 'a list): int = 
  match l with
    [] -> 0 |
    first::rest -> (count p rest) + (if p first then 1 else 0)
  
let rec exists_opt (f: 'a -> 'b option) (l: 'a list): 'b option =
  match l with 
    [] -> None |
    first::rest -> 
      let (r: 'b option) = f first in
      match r with
	None -> exists_opt f rest |
	Some v -> r

let rec partition_opt (f: 'a -> 'b option) (l: 'a list): ('b list * 'a list) =
  match l with
    [] -> ([], []) |
    first::rest -> 
      let (yes, no)= partition_opt f rest in
      match (f first) with
        None -> (yes, first::no) |
	Some v -> (v::yes, no)

let map_opt (f: 'a -> 'b) (a: 'a option): 'b option =
  match a with
  None -> (None: 'b option) |
  Some v -> Some (f v)

let rec filter_opt (l: 'a option list): 'a list = 
  match l with 
  [] -> [] |
  (None::rest) -> filter_opt rest |
  (Some first::rest) -> first::(filter_opt rest)

let is_something (opt:'a option): bool =
  match opt with			       
    None -> false |
    Some _ -> true

let de_opt (opt: 'a option): 'a = 
  match opt with
    None -> assert(false) |
    Some x -> x

let read_char c =
  let rc = ref ' ' in
  rc:= input_char stdin;
  while !rc==' ' do 
    rc:= input_char stdin
  done;
  !rc==c 

let rec drop_nth (l: 'a list) (k: int): 'a list =
  match l with 
    [] -> assert(false) | 
    first::rest -> 
      if k=0 then rest 
      else first::drop_nth rest (k-1)

(*
let read_list (read_element: unit -> 'a option): 'a list =
  let rec read_seq (): ('a list) = 
    let re= read_element () in begin
      ();
      if is_something re then
	(de_opt re)::(read_seq (): 'a list)
      else
	[]
    end
    if read_char ',' then read_seq ()
    else
      []
  in
  if read_char '[' then begin
    let l= read_seq () in
    if read_char ']' then l
    else []
  end
  else []

*)

let rec diff (l1:'a list) (l2: 'a list): 'a list =
  match l1 with
    [] -> [] |
    first::rest -> 
      if List.mem first l2 then diff rest l2
      else first::(diff rest l2)

let rec union (l1: 'a list) (l2: 'a list): 'a list =
  match l1 with
    [] -> l2 |
    first::rest -> 
      if List.mem first l2 then union rest l2
      else first::(union rest l2)

let find_index (p: 'a -> bool) (l: 'a list): int list =
  let rec fi l p n =
    match l with 
      [] -> [] |
      (h::t) -> 
	if (p h) then n::(fi t p (n+1))
	else fi t p (n+1)
  in fi l p 0

 let find_index2 (p: 'a  ->'b -> bool) (l1: 'a list) (l2: 'b list): (int * int) list =
    (* list of all pairs (n1, n2) s.t. p holds for the n1-th element of l1 and n2-th element of l2 *)
   let pair a l = List.map (function b -> (a, b)) l in
   let rec fi p l n =
     match l with 
       [] -> [] |
       (h::t) -> (pair n (find_index (p h) l2)) @ (fi p t (n+1))
   in fi p l1 0
   
let pair a b = (a, b)

let rec repeat x n =
  assert(n>=0);
  match n with
    0 -> []
  | _ -> x::(repeat x (n-1))

let rec cart_prod (l1: 'a list) (l2: 'b list) = 
  match l1 with 
    [] -> [] 
  | (hd::tl) -> (List.map (function x -> (hd, x)) l2) @ (cart_prod tl l2)

let array_exists_i (p: 'a -> int -> bool) (v: 'a array): int =
  let i= ref 0  in
  while !i<(Array.length v) && (not (p v.(!i) !i)) do
    i:= !i+1
  done;
  if !i>=(Array.length v) then -1 else !i

let array_exists (p: 'a -> bool) (v: 'a array): int =
  array_exists_i (fun a i -> p a) v

let array_exists2 (p: 'a -> 'b -> bool) (v1: 'a array) (v2: 'b array): int =
  assert((Array.length v1)=(Array.length v2));
  let i= ref 0  in
  while !i<(Array.length v1) && (not (p v1.(!i) v2.(!i))) do
    i:= !i+1
  done;
  if !i>=(Array.length v1) then -1 else !i

let array_find_index (p: 'a -> bool) (v: 'a array): int list =
  let result = ref [] in
  for i = 0 to (Array.length v)-1 do
    if p v.(i) then
      result:= i::!result;
  done;
  !result

let array_for_all_i (p: 'a -> int -> bool) (v: 'a array): int =
  let i= ref 0 in
  while !i<Array.length(v) && (p v.(!i) !i) do
    i:= !i+1
  done;
  if !i>=(Array.length v) then -1 else !i

let array_for_all (p: 'a -> bool) (v: 'a array): int =
  array_for_all_i (fun a i -> p a) v

let array_for_all2 (p: 'a -> 'b -> bool) (v1: 'a array) (v2: 'b array): int =
  let i= ref 0 in
  while !i<(Array.length v1) && (p v1.(!i) v2.(!i)) do
    i:= !i+1
  done;
  if !i>=(Array.length v1) then -1 else !i
    
let rec array_count (p: 'a -> bool) (a: 'a array): int = 
  let inc_true k a = if (p a) then k+1 else k in
  Array.fold_left inc_true 0 a 

let array_max_el (f: 'a -> float) (l: 'a array): (int * float) = 
  let max_el= ref (-1) 
  and max_val= ref neg_infinity 
  and i= ref 0 in
  while !i<(Array.length l) && !max_val<infinity do
    let new_val= (f l.(!i)) in
    if new_val > !max_val then begin
      max_el:= !i;
      max_val:= new_val
    end;
    i:= !i+1
  done;
  (!max_el, !max_val)

let is_a_number (x: float): bool=
  match classify_float x with 
    FP_normal | FP_subnormal | FP_zero -> true
  | _ -> false 

let matrix_init d1 d2 f =
  Array.init d1 (function i -> Array.init d2 (f i) )

let vector_prod (u: float array) (v: float array) =
  assert((Array.length u)=(Array.length v));
  let init_f i = u.(i)  *. v.(i) in
  Array.fold_left (+.) 0.0 (Array.init (Array.length u) init_f)

let vector_sum (u: float array) (v: float array) =
  assert((Array.length u)=(Array.length v));
  let init_f i = u.(i)  +. v.(i) in
  Array.init (Array.length u) init_f

let vector_diff (u: float array) (v: float array) =
  assert((Array.length u)=(Array.length v));
  let init_f i = u.(i) -. v.(i) in
  Array.init (Array.length u) init_f

let vector_scalar_mult (r: float) (u: float array)= 
  Array.init (Array.length u) (function i -> r *. u.(i))

let transpose (m: 'a array array): 'a array array =
  let init_f i = 
    Array.init (Array.length m) (function j -> m.(j).(i)) in
  Array.init (Array.length m.(0)) init_f 

let rec int_exists (lo: int) (hi: int) (f: int -> bool): int =
  if lo>hi then -1
  else if (f lo) then lo
  else int_exists (lo+1) hi f

let int_forall (lo: int) (hi: int) (f: int -> bool): int = 
  int_exists lo hi (function i -> not (f i))

